#include "config.h"
#include "libwtref.h"
#include "exc.h"
#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>

#if defined(PATH_MAX)
#  define LIBWTREF_PATH_MAX PATH_MAX
#elif defined(_MAX_PATH)
#  define LIBWTREF_PATH_MAX _MAX_PATH
#else
#  define LIBWTREF_PATH_MAX 4096
#endif

static void wtioc_writestdout_x(struct wtr_context* const ctx, const char* p,
	size_t n)
{
	fwrite(p, n, 1, stdout);
	wtrContextPushReturnValues(ctx, 0, NULL);
}

static void wtioc_writestdout(struct wtr_vm* const vm,
	struct wtr_context* const ctx, wt_vmptr p, size_t n)
{
	char buf[256];

	FILE* const f = stdout;

	while(sizeof(buf)<n)
	{
		if(WTE_OK!=wtrVMReadMemory(vm, p, buf, sizeof(buf)))
		{
			n = 0;
			break;
		}

		fwrite(buf, sizeof(buf), 1, f);
		p += sizeof(buf);
		n -= sizeof(buf);
	}

	if(0<n && WTE_OK==wtrVMReadMemory(vm, p, buf, n))
	{
		fwrite(buf, n, 1, f);
	}

	wtrContextPushReturnValues(ctx, 0, NULL);
}

static void wtioc_freadln(struct wtr_context* const ctx, FILE* const file,
	char* const buf, wt_u32 size, const wt_u8 sep)
{
	int c;
	char* ptr = buf;

	if(EOF==(c=fgetc(file)))
	{
		/* we read EOF first, EndOfFileException must be raised */
		wtrContextRaiseException(ctx, libwtref_exc_EndOfFileException);
		return;
	}

	do
	{
		if(sep==(wt_u8)c)
		{
			break;
		}

		if(0==size)
		{
			/* ran out of buffer space, must raise IOException */
			wtrContextRaiseException(ctx, libwtref_exc_IOException);
			return;
		}

		--size;
		*ptr++ = c;
	}
	while(EOF!=(c=fgetc(file)));

	/* we can reach this point by reaching EOF or hitting the separator -- */
	/* either way, send in the read number of character */
	{
		wt_value v_len;
		v_len.writable_type = WTRT_TYPEINFO(0, WTTYPE_U32);
		v_len.v_u32 = (wt_u32)(ptr - buf);
		wtrContextPushReturnValues(ctx, 1, &v_len);
	}
}

static void wtioc_access(struct wtr_vm* const vm,
	struct wtr_context* const ctx, wt_vmptr vma, wt_u32 len, wt_u32 mode)
{
#ifndef LIBWTREF_ACCESS_F_OK
#  define LIBWTREF_ACCESS_F_OK 0
#endif /* LIBWTREF_ACCESS_F_OK */
#ifndef LIBWTREF_ACCESS_W_OK
#  define LIBWTREF_ACCESS_W_OK 2
#endif /* LIBWTREF_ACCESS_W_OK */
#ifndef LIBWTREF_ACCESS_R_OK
#  define LIBWTREF_ACCESS_R_OK 4
#endif /* LIBWTREF_ACCESS_R_OK */
#ifndef LIBWTREF_ACCESS_X_OK
#  define LIBWTREF_ACCESS_X_OK 0
#endif /* LIBWTREF_ACCESS_X_OK */

	int amode = LIBWTREF_ACCESS_F_OK;
	wt_value val;

	char pathname[LIBWTREF_PATH_MAX];

	if(sizeof(pathname)<=len ||
		WTE_OK!=wtrVMReadMemory(vm, vma, pathname, len))
	{
		wtrContextRaiseException(ctx, libwtref_exc_IOException);
		return;
	}

	pathname[len] = 0;

	if(WTFA_READ & mode) { amode |= LIBWTREF_ACCESS_R_OK; }
	if(WTFA_WRITE & mode) { amode |= LIBWTREF_ACCESS_W_OK; }
	if(WTFA_EXECUTE & mode) { amode |= LIBWTREF_ACCESS_X_OK; }

	val.writable_type = WTRT_TYPEINFO(0, WTTYPE_BOOL);
	val.v_bool = (0==access(pathname, amode));
	wtrContextPushReturnValues(ctx, 1, &val);
}

static void push_stat(struct wtr_context* const ctx, struct stat* const st,
	const int res)
{
	wt_value val[5];

	val[0].writable_type = WTRT_TYPEINFO(0, WTTYPE_BOOL);
	val[1].writable_type = WTRT_TYPEINFO(0, WTTYPE_CHAR);
	val[2].writable_type = WTRT_TYPEINFO(0, WTTYPE_U64);
	val[3].writable_type = WTRT_TYPEINFO(0, WTTYPE_U64);
	val[4].writable_type = WTRT_TYPEINFO(0, WTTYPE_U64);

	if(0==res)
	{
		char wtdt;

		val[0].v_bool = WT_TRUE;

#ifdef LIBWTREF_STAT_ISBLK
		if(LIBWTREF_STAT_ISBLK(st->st_mode))
		{
			wtdt = WTDT_BLK;
		}
		else
#endif /* LIBWTREF_STAT_ISBLK */
#ifdef LIBWTREF_STAT_ISCHR
		if(LIBWTREF_STAT_ISCHR(st->st_mode))
		{
			wtdt = WTDT_CHR;
		}
		else
#endif /* LIBWTREF_STAT_ISCHR */
#ifdef LIBWTREF_STAT_ISDIR
		if(LIBWTREF_STAT_ISDIR(st->st_mode))
		{
			wtdt = WTDT_DIR;
		}
		else
#endif /* LIBWTREF_STAT_ISDIR */
#ifdef LIBWTREF_STAT_ISFIFO
		if(LIBWTREF_STAT_ISFIFO(st->st_mode))
		{
			wtdt = WTDT_FIFO;
		}
		else
#endif /* LIBWTREF_STAT_ISFIFO */
#ifdef LIBWTREF_STAT_ISLNK
		if(LIBWTREF_STAT_ISLNK(st->st_mode))
		{
			wtdt = WTDT_LNK;
		}
		else
#endif /* LIBWTREF_STAT_ISLNK */
#ifdef LIBWTREF_STAT_ISREG
		if(LIBWTREF_STAT_ISREG(st->st_mode))
		{
			wtdt = WTDT_REG;
		}
		else
#endif /* LIBWTREF_STAT_ISREG */
#ifdef LIBWTREF_STAT_ISSOCK
		if(LIBWTREF_STAT_ISSOCK(st->st_mode))
		{
			wtdt = WTDT_SOCK;
		}
		else
#endif /* LIBWTREF_STAT_ISSOCK */
		{
			wtdt = WTDT_UNKNOWN;
		}

		val[1].v_char = wtdt;

		val[2].v_u64 = st->st_size;

#if defined(LIBWTREF_STAT_HAVE_ST_MTIM)
		val[3].v_u64 =
			(((wt_u64)st->st_mtim.tv_sec)*1000)
			+
			((wt_u64)(st->st_mtim.tv_nsec/(1000*1000)))
			;
#elif defined(LIBWTREF_STAT_HAVE_ST_MTIME)
		val[3].v_u64 = ((wt_u64)st->st_mtime)*1000;
#else
		val[3].v_u64 = 0;
#endif

#if defined(LIBWTREF_STAT_HAVE_ST_BIRTHTIM)
		val[4].v_u64 =
			(((wt_u64)st->st_birthtim.tv_sec)*1000)
			+
			((wt_u64)(st->st_birthtim.tv_nsec/(1000*1000)))
			;
#elif defined(LIBWTREF_STAT_HAVE_ST_BIRTHTIME)
		val[4].v_u64 = ((wt_u64)st->st_birthtime)*1000;
#else
		val[4].v_u64 = 0;
#endif
	}
	else
	{
		val[0].v_bool = WT_FALSE;
		val[1].v_char = 0;
		val[2].v_u64  = 0;
		val[3].v_u64  = 0;
		val[4].v_u64  = 0;
	}

	wtrContextPushReturnValues(ctx, sizeof(val)/sizeof(val[0]), val);
}

static void wtioc_stat(struct wtr_vm* const vm, struct wtr_context* const ctx,
	wt_vmptr vma, wt_u32 len)
{
	struct stat st;

	char pathname[LIBWTREF_PATH_MAX];

	if(sizeof(pathname)<=len ||
		WTE_OK!=wtrVMReadMemory(vm, vma, pathname, len))
	{
		wtrContextRaiseException(ctx, libwtref_exc_IOException);
		return;
	}

	pathname[len] = 0;

	push_stat(ctx, &st, stat(pathname, &st));
}

static void wtioc_fstat(struct wtr_context* const ctx, wt_file file)
{
	struct stat st;
	push_stat(ctx, &st, fstat(LIBWTREF_FILENO_CALL((FILE*)file), &st));
}

static void wtioc_fopen(struct wtr_vm* const vm,
	struct wtr_context* const ctx, wt_vmptr vma, wt_u32 nlen, char mode)
{
	const char* smode;
	FILE* file;
	char pathname[LIBWTREF_PATH_MAX];

	switch(mode)
	{
	case 'r': smode = "rb";  break;
	case 'R': smode = "r+b"; break;
	case 'w': smode = "wb";  break;
	case 'W': smode = "w+b"; break;
	case 'a': smode = "ab";  break;
	case 'A': smode = "a+b"; break;
	default: goto io_exc;
	}

	if(sizeof(pathname)<=nlen ||
		WTE_OK!=wtrVMReadMemory(vm, vma, pathname, nlen))
	{
		goto io_exc;
	}

	pathname[nlen] = 0;

	if(NULL!=(file=fopen(pathname, smode)))
	{
		wt_value val;
		val.writable_type = WTRT_TYPEINFO(0, WTTYPE_FILE);
		val.v_file = (wt_file)file;
		wtrContextPushReturnValues(ctx, 1, &val);
		return;
	}

io_exc:;
	wtrContextRaiseException(ctx, libwtref_exc_IOException);
}

static void wtioc_fclose(struct wtr_context* const ctx, wt_file file)
{
	fclose((FILE*)file);
	wtrContextPushReturnValues(ctx, 0, NULL);
}

static void wtioc_fseek(struct wtr_context* const ctx, wt_file file,
	wt_i64 offset, wt_u32 whence)
{
	int w;
	wt_value val;

	switch(whence)
	{
	case WTSW_SET: w = SEEK_SET; break;
	case WTSW_CUR: w = SEEK_CUR; break;
	case WTSW_END: w = SEEK_END; break;

	default:
	io_exc:;
		wtrContextRaiseException(ctx, libwtref_exc_IOException);
		return;
	}

	val.writable_type = WTRT_TYPEINFO(0, WTTYPE_I64);

	if(0!=fseek((FILE*)file, offset, w))
	{
		goto io_exc;
	}

	val.v_i64 = ftell((FILE*)file);
	clearerr((FILE*)file);

	wtrContextPushReturnValues(ctx, 1, &val);
}

static void wtioc_feof(struct wtr_context* const ctx, wt_file file)
{
	wt_value val;

	val.writable_type = WTRT_TYPEINFO(0, WTTYPE_BOOL);
	val.v_bool = !!feof((FILE*)file);

	wtrContextPushReturnValues(ctx, 1, &val);
}

static void wtioc_fwrite(struct wtr_vm* const vm,
	struct wtr_context* const ctx, wt_file file, wt_vmptr vma, wt_u32 size)
{
	wt_value val;
	unsigned char buf[256];

	val.writable_type = WTRT_TYPEINFO(0, WTTYPE_U32);
	val.v_u32 = 0;

	while(sizeof(buf)<size)
	{
		size_t n;

		if(WTE_OK!=wtrVMReadMemory(vm, vma, buf, sizeof(buf)))
		{
			goto leave;
		}

		n = fwrite(buf, 1, sizeof(buf), (FILE*)file);

		val.v_u32 += n;
		vma += n;
		size -= n;

		if(sizeof(buf)>n)
		{
			goto leave;
		}
	}

	if(0<size)
	{
		if(WTE_OK!=wtrVMReadMemory(vm, vma, buf, size))
		{
			goto leave;
		}

		val.v_u32 += fwrite(buf, 1, size, (FILE*)file);
	}

leave:;
	wtrContextPushReturnValues(ctx, 1, &val);
}

static void wtioc_fwrite_x(struct wtr_context* const ctx, wt_file file,
	const char* buf, wt_u32 size)
{
	wt_value val;

	val.writable_type = WTRT_TYPEINFO(0, WTTYPE_U32);
	val.v_u32 = fwrite(buf, 1, size, (FILE*)file);

	wtrContextPushReturnValues(ctx, 1, &val);
}

static void wtioc_fread(struct wtr_vm* const vm,
	struct wtr_context* const ctx, wt_file file, wt_vmptr vma, wt_u32 size)
{
	wt_value val;
	unsigned char buf[256];

	val.writable_type = WTRT_TYPEINFO(0, WTTYPE_U32);
	val.v_u32 = 0;

	while(sizeof(buf)<size)
	{
		size_t n = fread(buf, 1, sizeof(buf), (FILE*)file);

		if(WTE_OK!=wtrVMWriteMemory(vm, vma, buf, n))
		{
			goto leave;
		}

		val.v_u32 += n;
		vma += n;
		size -= n;

		if(sizeof(buf)>n)
		{
			goto leave;
		}
	}

	if(0<size)
	{
		size_t n = fread(buf, 1, size, (FILE*)file);

		if(WTE_OK!=wtrVMWriteMemory(vm, vma, buf, n))
		{
			goto leave;
		}

		val.v_u32 += n;
	}

leave:;
	wtrContextPushReturnValues(ctx, 1, &val);
}

#if LIBWTREF_DIR_READDIR

struct dir_impl
{
	DIR*           dir;
	struct dirent* ent;
	size_t         nlen;
	char           typ;
};

static void wtioc_opendir(struct wtr_vm* const vm,
	struct wtr_context* const ctx, wt_vmptr vma, wt_u32 nlen)
{
	DIR* dir;
	struct dir_impl* di;
	wt_value val;

	char pathname[LIBWTREF_PATH_MAX];

	if(sizeof(pathname)<=nlen ||
		WTE_OK!=wtrVMReadMemory(vm, vma, pathname, nlen))
	{
		wtrContextRaiseException(ctx, libwtref_exc_IOException);
		return;
	}

	pathname[nlen] = 0;

	if(NULL==(dir=opendir(pathname)))
	{
	io_exc:;
		wtrContextRaiseException(ctx, libwtref_exc_IOException);
		return;
	}

	if(NULL==(di=(struct dir_impl*)malloc(sizeof(struct dir_impl))))
	{
		closedir(dir);
		goto io_exc;
	}

	di->dir = dir;
	di->ent = NULL;

	val.writable_type = WTRT_TYPEINFO(0, WTTYPE_DIR);
	val.v_dir = (wt_dir)di;

	wtrContextPushReturnValues(ctx, 1, &val);
}

static void wtioc_closedir(struct wtr_context* const ctx, wt_dir dir)
{
	struct dir_impl* const di = (struct dir_impl*)dir;

	closedir(di->dir);
	free(di);

	wtrContextPushReturnValues(ctx, 0, NULL);
}

static void wtioc_readdir(struct wtr_context* const ctx, wt_dir dir)
{
	struct dir_impl* const di = (struct dir_impl*)dir;
	wt_value val;

	di->ent = readdir(di->dir);

	val.writable_type = WTRT_TYPEINFO(0, WTTYPE_BOOL);

	if(NULL!=di->ent)
	{
		val.v_bool = WT_TRUE;

		di->nlen = strlen(di->ent->d_name);

		switch(di->ent->d_type)
		{
		case DT_BLK:  di->typ = WTDT_BLK;     break;
		case DT_CHR:  di->typ = WTDT_CHR;     break;
		case DT_DIR:  di->typ = WTDT_DIR;     break;
		case DT_FIFO: di->typ = WTDT_FIFO;    break;
		case DT_LNK:  di->typ = WTDT_LNK;     break;
		case DT_REG:  di->typ = WTDT_REG;     break;
		case DT_SOCK: di->typ = WTDT_SOCK;    break;
		default:      di->typ = WTDT_UNKNOWN; break;
		}
	}
	else
	{
		val.v_bool = WT_FALSE;
	}

	wtrContextPushReturnValues(ctx, 1, &val);
}

static void wtioc_dirgettype(struct wtr_context* const ctx, wt_dir dir)
{
	struct dir_impl* const di = (struct dir_impl*)dir;

	if(NULL!=di->ent)
	{
		wt_value val;

		val.writable_type = WTRT_TYPEINFO(0, WTTYPE_CHAR);
		val.v_char = di->typ;

		wtrContextPushReturnValues(ctx, 1, &val);

		return;
	}

	wtrContextRaiseException(ctx, libwtref_exc_IOException);
}

static void wtioc_dirgetname(struct wtr_context* const ctx, wt_dir dir,
	char* buf, wt_u32 size)
{
	struct dir_impl* const di = (struct dir_impl*)dir;

	if(NULL!=di->ent && size>=di->nlen)
	{
		wt_value val;

		size = di->nlen;

		val.writable_type = WTRT_TYPEINFO(0, WTTYPE_U32);
		val.v_u32 = size;

		memcpy(buf, di->ent->d_name, size);

		wtrContextPushReturnValues(ctx, 1, &val);

		return;
	}

	wtrContextRaiseException(ctx, libwtref_exc_IOException);
}

#elif LIBWTREF_DIR_FINDNEXTFILE

struct dir_impl
{
	HANDLE           find;
	WIN32_FIND_DATAA data;
	int              first;
};

static void wtioc_opendir(struct wtr_vm* const vm,
	struct wtr_context* const ctx, wt_vmptr vma, wt_u32 nlen)
{
	char find_path[LIBWTREF_PATH_MAX+4];

	if(LIBWTREF_PATH_MAX<=nlen ||
		WTE_OK!=wtrVMReadMemory(vm, vma, pathname, nlen))
	{
		wtrContextRaiseException(ctx, libwtref_exc_IOException);
		return;
	}

	strcpy(find_path+nlen, "/*");

	{
		struct dir_impl* const di =
			(struct dir_impl*)malloc(sizeof(struct dir_impl));

		if(NULL!=di)
		{
			strcpy(find_path, name);
			strcat(find_path, "/*");

			if(INVALID_HANDLE_VALUE!=(di->find=FindFirstFileA(find_path,
				&di->data)))
			{
				wt_value val;

				di->first = 1;

				val.writable_type = WTRT_TYPEINFO(0, WTTYPE_DIR);
				val.v_dir = (wt_dir)di;

				wtrContextPushReturnValues(ctx, 1, &val);
				return;
			}
			else
			{
				free(di);
			}
		}
	}

	wtrContextRaiseException(ctx, libwtref_exc_IOException);
}

static void wtioc_closedir(struct wtr_context* const ctx, wt_dir dir)
{
	struct dir_impl* const di = (struct dir_impl*)dir;

	FindClose(di->find);
	free(di);

	wtrContextPushReturnValues(ctx, 0, NULL);
}

static void wtioc_readdir(struct wtr_context* const ctx, wt_dir dir)
{
	struct dir_impl* const di = (struct dir_impl*)dir;
	wt_value val;

	val.writable_type = WTRT_TYPEINFO(0, WTTYPE_BOOL);

	if(di->first)
	{
		di->first = 0;
		val.v_bool = WT_TRUE;
	}
	else
	if(FindNextFileA(di->find, &di->data))
	{
		val.v_bool = WT_TRUE;
	}
	else
	{
		val.v_bool = WT_FALSE;
	}

	wtrContextPushReturnValues(ctx, 1, &val);
}

static void wtioc_dirgettype(struct wtr_context* const ctx, wt_dir dir)
{
	struct dir_impl* const di = (struct dir_impl*)dir;
	wt_value val;

	if(di->first)
	{
		/* no readdir() called yet */
		wtrContextRaiseException(ctx, libwtref_exc_IOException);
		return;
	}

	val.writable_type = WTRT_TYPEINFO(0, WTTYPE_CHAR);

	if(FILE_ATTRIBUTE_DIRECTORY & di->data.dwFileAttributes)
	{
		val.v_char = WTDT_DIR;
	}
	else
	{
		val.v_char = WTDT_REG;
	}

	wtrContextPushReturnValues(ctx, 1, &val);
}

static void wtioc_dirgetname(struct wtr_context* const ctx, wt_dir dir,
	char* buf, wt_u32 size)
{
	struct dir_impl* const di = (struct dir_impl*)dir;

	if(!di->first)
	{
		const size_t nlen = strlen(di->data.cFileName);

		if(nlen<=size)
		{
			wt_value val;

			val.writable_type = WTRT_TYPEINFO(0, WTTYPE_U32);
			val.v_u32 = nlen;

			memcpy(buf, di->data.cFileName, nlen);

			wtrContextPushReturnValues(ctx, 1, &val);

			return;
		}
	}

	wtrContextRaiseException(ctx, libwtref_exc_IOException);
}

#else

static void wtioc_opendir(struct wtr_vm* const vm,
	struct wtr_context* const ctx, wt_vmptr vma, wt_u32 nlen)
{
	wtrContextRaiseException(ctx, libwtref_exc_IOException);
}

static void wtioc_closedir(struct wtr_context* const ctx, wt_dir dir)
{
	wtrContextRaiseException(ctx, libwtref_exc_IOException);
}

static void wtioc_readdir(struct wtr_context* const ctx, wt_dir dir)
{
	wtrContextRaiseException(ctx, libwtref_exc_IOException);
}

static void wtioc_dirgettype(struct wtr_context* const ctx, wt_dir dir)
{
	wtrContextRaiseException(ctx, libwtref_exc_IOException);
}

static void wtioc_dirgetname(struct wtr_context* const ctx, wt_dir dir,
	char* buf, wt_u32 size)
{
	wtrContextRaiseException(ctx, libwtref_exc_IOException);
}

#endif

void libwtref_handle_iocall(struct wtr_vm* const vm,
	struct wtr_context* const ctx, LIBWTREF_PFN_ERROR_CB errcb,
	void* errcb_priv)
{
	wt_error        err;
	enum wtr_iocall iocall;
	wt_u32          argc;
	wt_value        argv[WTR_IOCALL_ARGMAX];

	err = wtrContextGetIOCallParams(ctx, &iocall, &argc, WTR_IOCALL_ARGMAX,
		argv);

	if(WTE_OK!=err)
	{
		errcb(errcb_priv, "Could not get I/O call parameters", err);
		wtrContextRaiseException(ctx, libwtref_exc_IOException);
		return;
	}

	switch(iocall)
	{
	case WTIOC_WRITESTDOUT:
		if(WTRT_TYPEINFO(0, WTTYPE_XSTRING)==argv[0].type)
		{
			wtioc_writestdout_x(ctx, argv[0].v_xstring.p,
				argv[0].v_xstring.n);
		}
		else
		{
			wtioc_writestdout(vm, ctx, argv[0].v_string.p,
				argv[0].v_string.n);
		}
		break;

	case WTIOC_READLNSTDIN:
		wtioc_freadln(ctx, stdin, argv[0].v_xstring.p, argv[0].v_xstring.n,
			argv[1].v_char);
		break;

	case WTIOC_ACCESS:
		wtioc_access(vm, ctx, argv[0].v_string.p, argv[0].v_string.n,
			argv[1].v_u32);
		break;

	case WTIOC_STAT:
		wtioc_stat(vm, ctx, argv[0].v_string.p, argv[0].v_string.n);
		break;

	case WTIOC_FSTAT:
		wtioc_fstat(ctx, argv[0].v_file);
		break;

	case WTIOC_FOPEN:
		wtioc_fopen(vm, ctx, argv[0].v_string.p, argv[0].v_string.n,
			argv[1].v_char);
		break;

	case WTIOC_FCLOSE:
		wtioc_fclose(ctx, argv[0].v_file);
		break;

	case WTIOC_FSEEK:
		wtioc_fseek(ctx, argv[0].v_file, argv[1].v_i64, argv[2].v_u32);
		break;

	case WTIOC_FEOF:
		wtioc_feof(ctx, argv[0].v_file);
		break;

	case WTIOC_FWRITE:
		if(WTRT_TYPEINFO(0, WTTYPE_STRING)==argv[1].type)
		{
			wtioc_fwrite(vm, ctx, argv[0].v_file, argv[1].v_string.p,
				argv[1].v_string.n);
		}
		else
		{
			wtioc_fwrite_x(ctx, argv[0].v_file, argv[1].v_xstring.p,
				argv[1].v_xstring.n);
		}
		break;

	case WTIOC_FREAD:
		wtioc_fread(vm, ctx, argv[0].v_file, argv[1].v_string.p,
			argv[1].v_string.n);
		break;

	case WTIOC_FREADLN:
		wtioc_freadln(ctx, (FILE*)argv[0].v_file, argv[1].v_xstring.p,
			argv[1].v_xstring.n, argv[2].v_char);
		break;

	case WTIOC_OPENDIR:
		wtioc_opendir(vm, ctx, argv[0].v_string.p, argv[0].v_string.n);
		break;

	case WTIOC_CLOSEDIR:
		wtioc_closedir(ctx, argv[0].v_dir);
		break;

	case WTIOC_READDIR:
		wtioc_readdir(ctx, argv[0].v_dir);
		break;

	case WTIOC_DIRGETTYPE:
		wtioc_dirgettype(ctx, argv[0].v_dir);
		break;

	case WTIOC_DIRGETNAME:
		wtioc_dirgetname(ctx, argv[0].v_dir, argv[1].v_xstring.p,
			argv[1].v_xstring.n);
		break;

	default:
		/* anything else is IOException */
		wtrContextRaiseException(ctx, libwtref_exc_IOException);
		break;
	}
}
