#ifndef WTRUNTIME_VM_H
#define WTRUNTIME_VM_H

#include "config.h"
#include "types.h"
#include "wtruntime.h"
#include "vector.h"
#include "gidtree.h"
#include <libvmma/vmmu32.h>
#include <stddef.h>

struct wtr_program;
struct wtr_type;
struct wtr_symtab_b;
struct buddy32v;

struct wtr_vm
{
	struct vmmu32          vmmu;            /* virtual memory management unit */
	struct buddy32v*       heap_buddy;      /* virtual buddy allocator for heap */
	wt_uint64_t            next_prog_id;    /* next allocatable program ID */
	wt_u32                 num_context;     /* number of created context structures */
	wt_u32                 external_tag;    /* tag of the wt_pfn that holds an external C function */
	wt_u32                 accepted_flags;  /* flags accepted during the initialization (WTRPF_*) */
	wt_u32                 pc_tcl_ja;       /* absolute PC of the "TCL ja.s" routine */
	wt_u32                 pc_tcl_leave;    /* absolute PC of the context cleanup ("TCL leave") routine */
	wt_u32                 pc_strd_leave;   /* absolute PC of the "STRD leave" routine */
	wt_u32                 pc_cfp_leave;    /* absolute PC of the "CFP leave" routine */
	wt_u32                 ratio_css;       /* CSS ratio from 'stack_size' */
	wt_u32                 ratio_dss;       /* DSS ratio from 'stack_size' */
	void*                  alloc_ctx;       /* memory allocator context */
	WT_PFN_ALLOC           alloc_func;      /* memory allocator function */
	void*                  user_ptr;        /* an additional user pointer, this is opaque to the VM */
	void*                  heap_sysmem;     /* system memory allocated for heap buddy state */
	wt_u32*                vma_segacq;      /* VMMU segment acquire counters (larger than builtin counters) */
	wt_u32                 vma_segsh;       /* VMMU segment size shift (log2) */
	struct wtr_gidtree     gidtree;         /* GID-tree to translate GID to symbol */
	struct wtr_list        list_programs;   /* list of all loaded programs */
	struct wtr_list        list_symtabs;    /* list of all symbol tables */
	struct wtr_symtab_b*   symtab_b;        /* table for builtin and user symbols */
	wt_u32                 xs_x;            /* xorshift prng state */
	wt_u32                 xs_y;            /* xorshift prng state */
	wt_u32                 xs_z;            /* xorshift prng state */
	wt_u32                 xs_w;            /* xorshift prng state */
	void*                  loaderops_ctx;   /* unresolved object handlers */
	const struct wtr_loaderops* loaderops;  /* unresolved object handlers */
	const struct wtr_hashalg*   hashalgs;   /* hash algoritm array */
};

/* find the specified symbol in the loaded symbol tables */
struct wtr_sym* wtr_vm_find_sym(struct wtr_vm* const vm,
	struct wtr_sym* const sym);


wt_uint32_t wtr_vm_xorshift(struct wtr_vm* const vm);
void wtr_vm_xorshift_seed(struct wtr_vm* const vm, wt_uint32_t seed);


/* determine the size of a type */
size_t wtr_sizeof(struct wtr_vm* const vm, wt_typeid type);


/* copy memory regions from VMA to VMA (returns zero on success) */
int wtr_vma_memcpy(struct wtr_vm* const vm, wt_vmptr dst, wt_vmptr src,
	wt_u32 size);

/* copy overlapping memory regions from VMA to VMA (returns zero on success) */
int wtr_vma_memmove(struct wtr_vm* const vm, wt_vmptr dst, wt_vmptr src,
	wt_u32 size);

/* set memory in VMA (returns zero on success) */
int wtr_vma_memset(struct wtr_vm* const vm, wt_vmptr vma, int c,
	wt_u32 size);

/* compare memory regions in VMA (returns zero on success) */
int wtr_vma_memcmp(struct wtr_vm* const vm, wt_vmptr s1, wt_vmptr s2,
	wt_u32 size);

/* computes C-string length (returns zero on success) */
int wtr_vma_strlen(struct wtr_vm* const vm, wt_vmptr str, wt_u32* const len);

/* write from VMA */
#define wtr_vma_read(_vm_, _vma_, _ptr_, _size_) \
	vmmu32_read(&(_vm_)->vmmu, _vma_, _ptr_, _size_)

/* write to VMA */
#define wtr_vma_write(_vm_, _vma_, _ptr_, _size_) \
	vmmu32_write(&(_vm_)->vmmu, _vma_, _ptr_, _size_)


/* acquire/release VMMU segments of the specified region */
int wtr_vma_acquire(struct wtr_vm* const vm, wt_vmptr vma, wt_u32 size);
int wtr_vma_release(struct wtr_vm* const vm, wt_vmptr vma, wt_u32 size);


/* allocate memory that is accessible in the VM */
wt_vmptr wtr_vm_vma_malloc(struct wtr_vm* const vm, size_t size);
void wtr_vm_vma_free(struct wtr_vm* const vm, wt_vmptr memory);

/* reallocate system memory */
void* wtr_vm_sys_realloc(struct wtr_vm* const vm, void* memory,
	const size_t old_size, const size_t new_size);
void* wtr_vm_sys_malloc(struct wtr_vm* const vm, size_t size);
void wtr_vm_sys_free(struct wtr_vm* const vm, void* memory);

#endif /* WTRUNTIME_VM_H */
